@@ -4,29 +4,42 @@
 #
 # Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
 
-import struct
+
 from abc import ABC, abstractmethod
 from decimal import Decimal
 from io import BytesIO
 from itertools import zip_longest
 from pathlib import Path
 from shutil import copyfileobj
-from zlib import decompress
+from typing import (
+    Any,
+    BinaryIO,
+    Callable,
+    NamedTuple,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 
 from PIL import Image, ImageCms
-from PIL.TiffTags import TAGS_V2 as TIFF_TAGS
 
 from pikepdf import (
     Array,
     Dictionary,
     Name,
     Object,
+    Pdf,
     PdfError,
     Stream,
     StreamDecodeLevel,
     String,
     jbig2,
 )
+from pikepdf._qpdf import Buffer
+from pikepdf.models import _transcoding
 
 
 class DependencyError(Exception):
@@ -41,17 +54,21 @@ class NotExtractableError(Exception):
     "Indicates that an image cannot be directly extracted."
 
 
+class HifiPrintImageNotTranscodableError(NotExtractableError):
+    "Image contains high fidelity printing information and cannot be extracted."
+
+
 class InvalidPdfImageError(Exception):
     "This image is not valid according to the PDF 1.7 specification."
 
 
-def array_str(value):
-    """Simplify pikepdf objects to array of str. Keep Streams intact."""
+def array_str(value: Union[Object, str, list]):
+    """Simplify pikepdf objects to array of str. Keep Streams and dictionaries intact."""
 
     def _array_str(item):
         if isinstance(item, (list, Array)):
             return [_array_str(subitem) for subitem in item]
-        elif isinstance(item, (Stream, bytes, int)):
+        elif isinstance(item, (Stream, Dictionary, bytes, int)):
             return item
         elif isinstance(item, (Name, str)):
             return str(item)
@@ -76,7 +93,12 @@ def dict_or_array_dict(value):
     raise NotImplementedError(value)
 
 
-def metadata_from_obj(obj, name, type_, default):
+T = TypeVar('T')
+
+
+def metadata_from_obj(
+    obj: Union[Dictionary, Stream], name: str, type_: Callable[[Any], T], default: T
+) -> Optional[T]:
     val = getattr(obj, name, default)
     try:
         return type_(val)
@@ -86,53 +108,67 @@ def metadata_from_obj(obj, name, type_, default):
     raise NotImplementedError('Metadata access for ' + name)
 
 
+class PaletteData(NamedTuple):
+    """Returns the color space and binary representation of the palette.
+
+    ``base_colorspace`` is typically ``"RGB"`` or ``"L"`` (for grayscale).
+
+    ``palette`` is typically 256 or 256*3=768 bytes, for grayscale and RGB color
+    respectively, with each unit/triplet being the grayscale/RGB triplet values.
+    """
+
+    base_colorspace: str
+    palette: bytes
+
+
 class PdfImageBase(ABC):
 
     SIMPLE_COLORSPACES = {'/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray'}
     MAIN_COLORSPACES = SIMPLE_COLORSPACES | {'/DeviceCMYK', '/CalCMYK', '/ICCBased'}
+    PRINT_COLORSPACES = {'/Separation', '/DeviceN'}
 
     @abstractmethod
-    def _metadata(self, name, type_, default):
-        ...
+    def _metadata(self, name: str, type_: Type[T], default: T) -> T:
+        """Get metadata for this image type."""
 
     @property
-    def width(self):
-        """Width of the image data in pixels"""
-        return self._metadata('Width', int, None)
+    def width(self) -> int:
+        """Width of the image data in pixels."""
+        return self._metadata('Width', int, 0)
 
     @property
-    def height(self):
-        """Height of the image data in pixels"""
-        return self._metadata('Height', int, None)
+    def height(self) -> int:
+        """Height of the image data in pixels."""
+        return self._metadata('Height', int, 0)
 
     @property
-    def image_mask(self):
-        """``True`` if this is an image mask"""
+    def image_mask(self) -> bool:
+        """``True`` if this is an image mask."""
         return self._metadata('ImageMask', bool, False)
 
     @property
-    def _bpc(self):
-        """Bits per component for this image (low-level)"""
-        return self._metadata('BitsPerComponent', int, None)
+    def _bpc(self) -> Optional[int]:
+        """Bits per component for this image (low-level)."""
+        return self._metadata('BitsPerComponent', int, 0)
 
     @property
     def _colorspaces(self):
-        """Colorspace (low-level)"""
+        """Colorspace (low-level)."""
         return self._metadata('ColorSpace', array_str, [])
 
     @property
     def filters(self):
-        """List of names of the filters that we applied to encode this image"""
+        """List of names of the filters that we applied to encode this image."""
         return self._metadata('Filter', array_str, [])
 
     @property
     def decode_parms(self):
-        """List of the /DecodeParms, arguments to filters"""
+        """List of the /DecodeParms, arguments to filters."""
         return self._metadata('DecodeParms', dict_or_array_dict, [])
 
     @property
-    def colorspace(self):
-        """PDF name of the colorspace that best describes this image"""
+    def colorspace(self) -> Optional[str]:
+        """PDF name of the colorspace that best describes this image."""
         if self.image_mask:
             return None  # Undefined for image masks
         if self._colorspaces:
@@ -142,37 +178,64 @@ class PdfImageBase(ABC):
                 subspace = self._colorspaces[1]
                 if isinstance(subspace, str) and subspace in self.MAIN_COLORSPACES:
                     return subspace
-                if isinstance(subspace, list) and subspace[0] == '/ICCBased':
+                if isinstance(subspace, list) and subspace[0] in (
+                    '/ICCBased',
+                    '/DeviceN',
+                ):
                     return subspace[0]
+            if self._colorspaces[0] == '/DeviceN':
+                return '/DeviceN'
+
         raise NotImplementedError(
             "not sure how to get colorspace: " + repr(self._colorspaces)
         )
 
     @property
-    def bits_per_component(self):
-        """Bits per component of this image"""
-        if self._bpc is None:
+    def bits_per_component(self) -> int:
+        """Bits per component of this image."""
+        if self._bpc is None or self._bpc == 0:
             return 1 if self.image_mask else 8
         return self._bpc
 
     @property
     @abstractmethod
-    def is_inline(self):
-        ...
+    def is_inline(self) -> bool:
+        """Is this an inline image?"""
 
     @property
     @abstractmethod
-    def icc(self):
-        ...
+    def icc(self) -> Optional[ImageCms.ImageCmsProfile]:
+        """Returns ICC profile for this image if one is defined."""
 
     @property
-    def indexed(self):
-        """``True`` if the image has a defined color palette"""
+    def indexed(self) -> bool:
+        """``True`` if the image has a defined color palette."""
         return '/Indexed' in self._colorspaces
 
+    def _colorspace_has_name(self, name):
+        try:
+            cs = self._colorspaces
+            if cs[0] == '/Indexed' and cs[1][0] == name:
+                return True
+            if cs[0] == name:
+                return True
+        except (IndexError, AttributeError, KeyError):
+            pass
+        return False
+
     @property
-    def size(self):
-        """Size of image as (width, height)"""
+    def is_device_n(self) -> bool:
+        """``True`` if image has a /DeviceN (complex printing) colorspace."""
+        return self._colorspace_has_name('/DeviceN')
+
+    @property
+    def is_separation(self) -> bool:
+        """``True`` if image has a /DeviceN (complex printing) colorspace."""
+        return self._colorspace_has_name('/Separation')
+
+    @property
+    def size(self) -> Tuple[int, int]:
+        """Size of image as (width, height)."""
         return self.width, self.height
 
     def _approx_mode_from_icc(self):
@@ -191,7 +254,7 @@ class PdfImageBase(ABC):
         return mode_from_xcolor_space.get(xcolor_space, '')
 
     @property
-    def mode(self):
+    def mode(self) -> str:
         """``PIL.Image.mode`` equivalent for this image, where possible
 
         If an ICC profile is attached to the image, we still attempt to resolve a Pillow
@@ -199,24 +262,27 @@ class PdfImageBase(ABC):
         """
 
         m = ''
-        if self.indexed:
+        if self.is_device_n:
+            m = 'DeviceN'
+        elif self.is_separation:
+            m = 'Separation'
+        elif self.indexed:
             m = 'P'
-        elif self.bits_per_component == 1:
+        elif self.colorspace == '/DeviceGray' and self.bits_per_component == 1:
             m = '1'
-        elif self.bits_per_component == 8:
-            if self.colorspace == '/DeviceRGB':
-                m = 'RGB'
-            elif self.colorspace == '/DeviceGray':
-                m = 'L'
-            elif self.colorspace == '/DeviceCMYK':
-                m = 'CMYK'
-            elif self.colorspace == '/ICCBased':
-                try:
-                    m = self._approx_mode_from_icc()
-                except (ValueError, TypeError) as e:
-                    raise NotImplementedError(
-                        "Not sure how to handle PDF image of this type"
-                    ) from e
+        elif self.colorspace == '/DeviceGray' and self.bits_per_component > 1:
+            m = 'L'
+        elif self.colorspace == '/DeviceRGB':
+            m = 'RGB'
+        elif self.colorspace == '/DeviceCMYK':
+            m = 'CMYK'
+        elif self.colorspace == '/ICCBased':
+            try:
+                m = self._approx_mode_from_icc()
+            except (ValueError, TypeError) as e:
+                raise NotImplementedError(
+                    "Not sure how to handle PDF image of this type"
+                ) from e
         if m == '':
             raise NotImplementedError(
                 "Not sure how to handle PDF image of this type"
@@ -238,12 +304,8 @@ class PdfImageBase(ABC):
         return list(zip_longest(self.filters, self.decode_parms, fillvalue={}))
 
     @property
-    def palette(self):
-        """Retrieves the color palette for this image
-
-        Returns:
-            tuple (base_colorspace: str, palette: bytes)
-        """
+    def palette(self) -> Optional[PaletteData]:
+        """Retrieves the color palette for this image if applicable."""
 
         if not self.indexed:
             return None
@@ -251,53 +313,78 @@ class PdfImageBase(ABC):
             _idx, base, _hival, lookup = self._colorspaces
         except ValueError as e:
             raise ValueError('Not sure how to interpret this palette') from e
-        if self.icc:
+        if self.icc or self.is_device_n or self.is_separation:
             base = str(base[0])
         else:
             base = str(base)
         lookup = bytes(lookup)
-        if not base in self.MAIN_COLORSPACES:
-            raise NotImplementedError("not sure how to interpret this palette")
+        if base not in self.MAIN_COLORSPACES and base not in self.PRINT_COLORSPACES:
+            raise NotImplementedError(f"not sure how to interpret this palette: {base}")
         if base == '/DeviceRGB':
             base = 'RGB'
         elif base == '/DeviceGray':
             base = 'L'
         elif base == '/DeviceCMYK':
             base = 'CMYK'
+        elif base == '/DeviceN':
+            base = 'DeviceN'
+        elif base == '/Separation':
+            base = 'Separation'
         elif base == '/ICCBased':
             base = self._approx_mode_from_icc()
-        return base, lookup
+        return PaletteData(base, lookup)
 
     @abstractmethod
-    def as_pil_image(self):
-        ...
+    def as_pil_image(self) -> Image.Image:
+        """Convert this PDF image to a Python PIL (Pillow) image."""
 
     @staticmethod
-    def _unstack_compression(buffer, filters):
-        """Remove stacked compression where it appears.
-
-        Stacked compression means when an image is set to:
-            ``[/FlateDecode /DCTDecode]``
-        for example.
-
-        Only Flate can be stripped off the front currently.
+    def _remove_simple_filters(obj, filters):
+        """Remove simple lossless compression where it appears.
 
         Args:
-            buffer (pikepdf._qpdf.Buffer): the compressed image data
+            obj (pikepdf.Stream): the compressed object
             filters (list of str): all files on the data
         """
-        data = memoryview(buffer)
-        while len(filters) > 1 and filters[0] == '/FlateDecode':
-            data = decompress(data)
-            filters = filters[1:]
-        return data, filters
+
+        COMPLEX_FILTERS = {
+            '/DCTDecode',
+            '/JPXDecode',
+            '/JBIG2Decode',
+            '/CCITTFaxDecode',
+        }
+
+        idx = [n for n, item in enumerate(filters) if item in COMPLEX_FILTERS]
+        if idx:
+            if len(idx) > 1:
+                raise NotImplementedError(
+                    f"Object {obj.objgen} has compound complex filters: {filters}. "
+                    "We cannot decompress this."
+                )
+            simple_filters = filters[: idx[0]]
+            complex_filters = filters[idx[0] :]
+        else:
+            simple_filters = filters
+            complex_filters = []
+
+        if not simple_filters:
+            return obj.read_raw_bytes(), complex_filters
+
+        original_filters = obj.Filter
+        try:
+            obj.Filter = Array([Name(s) for s in simple_filters])
+            data = obj.read_bytes(StreamDecodeLevel.specialized)
+        finally:
+            obj.Filter = original_filters
+
+        return data, complex_filters
 
 
 class PdfImage(PdfImageBase):
     """Support class to provide a consistent API for manipulating PDF images
 
-    The data structure for images inside PDFs is irregular and flexible,
-    making it difficult to work with without introducing errors for less
+    The data structure for images inside PDFs is irregular and complex,
+    making it difficult to use without introducing errors for less
     typical cases. This class addresses these difficulties by providing a
     regular, Pythonic API similar in spirit (and convertible to) the Python
     Pillow imaging library.
@@ -363,7 +450,7 @@ class PdfImage(PdfImageBase):
 
     @property
     def is_inline(self):
-        """``False`` for image XObject"""
+        """``False`` for image XObject."""
         return False
 
     @property
@@ -376,13 +463,11 @@ class PdfImage(PdfImageBase):
         raise NotImplementedError("Don't know how to find ICC stream for image")
 
     @property
-    def icc(self):
+    def icc(self) -> Optional[ImageCms.ImageCmsProfile]:
         """If an ICC profile is attached, return a Pillow object that describe it.
 
         Most of the information may be found in ``icc.profile``.
 
-        Returns:
-            PIL.ImageCms.ImageCmsProfile
         """
         if self.colorspace not in ('/ICCBased', '/Indexed'):
             return None
@@ -400,7 +485,7 @@ class PdfImage(PdfImageBase):
                     ) from e
         return self._icc
 
-    def _extract_direct(self, *, stream):
+    def _extract_direct(self, *, stream: BinaryIO) -> str:
         """Attempt to extract the image directly to a usable image file
 
         If there is no way to extract the image without decompressing or
@@ -408,10 +493,10 @@ class PdfImage(PdfImageBase):
         generated will vary.
 
         Args:
-            stream: Writable stream to write data to
+            stream: Writable file stream to write data to, e.g. an open file
         """
 
-        def normal_dct_rgb():
+        def normal_dct_rgb() -> bool:
             # Normal DCTDecode RGB images have the default value of
             # /ColorTransform 1 and are actually in YUV. Such a file can be
             # saved as a standard JPEG. RGB JPEGs without YUV conversion can't
@@ -421,16 +506,14 @@ class PdfImage(PdfImageBase):
             ct = self.filter_decodeparms[0][1].get('/ColorTransform', DEFAULT_CT_RGB)
             return self.mode == 'RGB' and ct == DEFAULT_CT_RGB
 
-        def normal_dct_cmyk():
+        def normal_dct_cmyk() -> bool:
             # Normal DCTDecode CMYKs have /ColorTransform 0 and can be saved.
             # There is a YUVK colorspace but CMYK JPEGs don't generally use it
             DEFAULT_CT_CMYK = 0
             ct = self.filter_decodeparms[0][1].get('/ColorTransform', DEFAULT_CT_CMYK)
             return self.mode == 'CMYK' and ct == DEFAULT_CT_CMYK
 
-        data, filters = self._unstack_compression(
-            self.obj.get_raw_stream_buffer(), self.filters
-        )
+        data, filters = self._remove_simple_filters(self.obj, self.filters)
 
         if filters == ['/CCITTFaxDecode']:
             if self.colorspace == '/ICCBased':
@@ -448,61 +531,82 @@ class PdfImage(PdfImageBase):
 
         raise NotExtractableError()
 
-    def _extract_transcoded(self):
-        im = None
-        if self.mode == 'RGB' and self.bits_per_component == 8:
-            # No point in accessing the buffer here, size qpdf decodes to 3-byte
-            # RGB and Pillow needs RGBX for raw access
-            data = self.read_bytes()
-            im = Image.frombytes('RGB', self.size, data)
-        elif self.mode in ('L', 'P') and self.bits_per_component == 8:
-            buffer = self.get_stream_buffer()
-            stride = 0  # tell Pillow to calculate stride from line width
-            ystep = 1  # image is top to bottom in memory
-            im = Image.frombuffer('L', self.size, buffer, "raw", 'L', stride, ystep)
-            if self.mode == 'P':
-                base_mode, palette = self.palette
-                if base_mode in ('RGB', 'L'):
-                    im.putpalette(palette, rawmode=base_mode)
-                else:
-                    raise NotImplementedError('palette with ' + base_mode)
-        elif self.bits_per_component == 1:
-            if self.filters and self.filters[0] == '/JBIG2Decode':
-                if not jbig2.jbig2dec_available():
-                    raise DependencyError(
-                        "jbig2dec - not installed or installed version is too old "
-                        "(older than version 0.15)"
-                    )
-                jbig2_globals_obj = self.filter_decodeparms[0][1].get('/JBIG2Globals')
-                im = jbig2.extract_jbig2(self.obj, jbig2_globals_obj)
-            else:
-                data = self.read_bytes()
-                im = Image.frombytes('1', self.size, data)
+    def _extract_transcoded_248bits(self) -> Image.Image:
+        """Extract an image when there are 2/4/8 bits packed in byte data."""
+
+        stride = 0  # tell Pillow to calculate stride from line width
+        scale = 0 if self.mode == 'L' else 1
+        if self.bits_per_component in (2, 4):
+            buffer, stride = _transcoding.unpack_subbyte_pixels(
+                self.read_bytes(), self.size, self.bits_per_component, scale
+            )
+        elif self.bits_per_component == 8:
+            buffer = cast(memoryview, self.get_stream_buffer())
         else:
-            raise UnsupportedImageTypeError(repr(self))
+            raise InvalidPdfImageError("BitsPerComponent must be 1, 2, 4, 8, or 16")
 
-        if self.mode == 'P' and self.bits_per_component == 1:
-            # Fix paletted 1-bit images
+        if self.mode == 'P' and self.palette is not None:
             base_mode, palette = self.palette
-            if base_mode == 'RGB' and palette != b'\x00\x00\x00\xff\xff\xff':
-                im = im.convert('P')
-                im.putpalette(palette, rawmode=base_mode)
-                gp = im.getpalette()
-                gp[765:768] = gp[3:6]  # work around Pillow bug
-                im.putpalette(gp)
-            elif base_mode == 'L' and palette != b'\x00\xff':
-                im = im.convert('P')
-                im.putpalette(palette, rawmode=base_mode)
-                gp = im.getpalette()
-                gp[255] = gp[1]  # work around Pillow bug
-                im.putpalette(gp)
+            im = _transcoding.image_from_buffer_and_palette(
+                buffer,
+                self.size,
+                stride,
+                self.bits_per_component,
+                base_mode,
+                palette,
+            )
+        else:
+            im = _transcoding.image_from_byte_buffer(buffer, self.size, stride)
+        return im
 
-        if self.colorspace == '/ICCBased':
+    def _extract_transcoded_1bit(self) -> Image.Image:
+        if self.filters and self.filters[0] == '/JBIG2Decode':
+            if not jbig2.jbig2dec_available():
+                raise DependencyError(
+                    "jbig2dec - not installed or installed version is too old "
+                    "(older than version 0.15)"
+                )
+            jbig2_globals_obj = self.filter_decodeparms[0][1].get('/JBIG2Globals')
+            im = jbig2.extract_jbig2(self.obj, jbig2_globals_obj)
+        else:
+            if self.mode in ('RGB', 'CMYK'):
+                raise UnsupportedImageTypeError("1-bit RGB and CMYK are not supported")
+            data = self.read_bytes()
+            im = Image.frombytes('1', self.size, data)
+
+        if self.palette is not None:
+            base_mode, palette = self.palette
+            im = _transcoding.fix_1bit_palette_image(im, base_mode, palette)
+
+        return im
+
+    def _extract_transcoded(self) -> Image.Image:
+        if self.mode in {'DeviceN', 'Separation'}:
+            raise HifiPrintImageNotTranscodableError()
+
+        if self.mode == 'RGB' and self.bits_per_component == 8:
+            # Cannot use the zero-copy .get_stream_buffer here, we have 3-byte
+            # RGB and Pillow needs RGBX.
+            im = Image.frombuffer(
+                'RGB', self.size, self.read_bytes(), 'raw', 'RGB', 0, 1
+            )
+        elif self.mode == 'CMYK' and self.bits_per_component == 8:
+            im = Image.frombuffer(
+                'CMYK', self.size, self.get_stream_buffer(), 'raw', 'CMYK', 0, 1
+            )
+        elif self.mode in ('L', 'P') and 2 <= self.bits_per_component <= 8:
+            im = self._extract_transcoded_248bits()
+        elif self.bits_per_component == 1:
+            im = self._extract_transcoded_1bit()
+        else:
+            raise UnsupportedImageTypeError(repr(self) + ", " + repr(self.obj))
+
+        if self.colorspace == '/ICCBased' and self.icc is not None:
             im.info['icc_profile'] = self.icc.tobytes()
 
         return im
 
-    def _extract_to_stream(self, *, stream):
+    def _extract_to_stream(self, *, stream: BinaryIO) -> str:
         """Attempt to extract the image to a stream
 
         If possible, the compressed data is extracted and inserted into
@@ -514,7 +618,7 @@ class PdfImage(PdfImageBase):
             stream: Writable stream to write data to
 
         Returns:
-            str: The file format extension
+            The file format extension.
         """
 
         try:
@@ -522,19 +626,28 @@ class PdfImage(PdfImageBase):
         except NotExtractableError:
             pass
 
+        im = None
         try:
             im = self._extract_transcoded()
-            if im:
+            if im.mode == 'CMYK':
+                im.save(stream, format='tiff', compression='tiff_adobe_deflate')
+                return '.tiff'
+            elif im:
                 im.save(stream, format='png')
                 return '.png'
         except PdfError as e:
-            if 'getStreamData called on unfilterable stream' in str(e):
+            if 'called on unfilterable stream' in str(e):
                 raise UnsupportedImageTypeError(repr(self)) from e
             raise
+        finally:
+            if im:
+                im.close()
 
         raise UnsupportedImageTypeError(repr(self))
 
-    def extract_to(self, *, stream=None, fileprefix=''):
+    def extract_to(
+        self, *, stream: Optional[BinaryIO] = None, fileprefix: str = ''
+    ) -> str:
         """Attempt to extract the image directly to a usable image file
 
         If possible, the compressed data is extracted and inserted into
@@ -547,6 +660,8 @@ class PdfImage(PdfImageBase):
         When saving the image to a file, use a temporary filename, and then
         rename the file to its final name based on the returned file extension.
 
+        Images might be saved as any of .png, .jpg, or .tiff.
+
         Examples:
 
             >>> im.extract_to(stream=bytes_io)
@@ -564,9 +679,6 @@ class PdfImage(PdfImageBase):
             If *fileprefix* was provided, then the fileprefix with the
             appropriate extension. If no *fileprefix*, then an extension
             indicating the file type.
-
-        Return type:
-            str
         """
 
         if bool(stream) == bool(fileprefix):
@@ -582,19 +694,22 @@ class PdfImage(PdfImageBase):
             copyfileobj(bio, target)
         return str(filepath)
 
-    def read_bytes(self, decode_level=StreamDecodeLevel.specialized):
-        """Decompress this image and return it as unencoded bytes"""
+    def read_bytes(
+        self, decode_level: StreamDecodeLevel = StreamDecodeLevel.specialized
+    ) -> bytes:
+        """Decompress this image and return it as unencoded bytes."""
         return self.obj.read_bytes(decode_level=decode_level)
 
-    def get_stream_buffer(self, decode_level=StreamDecodeLevel.specialized):
-        """Access this image with the buffer protocol"""
+    def get_stream_buffer(
+        self, decode_level: StreamDecodeLevel = StreamDecodeLevel.specialized
+    ) -> Buffer:
+        """Access this image with the buffer protocol."""
         return self.obj.get_stream_buffer(decode_level=decode_level)
 
-    def as_pil_image(self):
-        """Extract the image as a Pillow Image, using decompression as necessary
+    def as_pil_image(self) -> Image.Image:
+        """Extract the image as a Pillow Image, using decompression as necessary.
 
-        Returns:
-            PIL.Image.Image
+        Caller must close the image.
         """
         try:
             bio = BytesIO()
@@ -610,8 +725,8 @@ class PdfImage(PdfImageBase):
 
         return im
 
-    def _generate_ccitt_header(self, data, icc=None):
-        """Construct a CCITT G3 or G4 header from the PDF metadata"""
+    def _generate_ccitt_header(self, data: bytes, icc: Optional[bytes] = None) -> bytes:
+        """Construct a CCITT G3 or G4 header from the PDF metadata."""
         # https://stackoverflow.com/questions/2641770/
         # https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf
 
@@ -638,63 +753,16 @@ class PdfImage(PdfImageBase):
         photometry = 1 if black_is_one else 0
 
         img_size = len(data)
-        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H'
-
-        tag_keys = {tag.name: key for key, tag in TIFF_TAGS.items()}
-        ifd_struct = '<HHLL'
 
         if icc is None:
             icc = b''
 
-        ifds = []
-
-        def header_length(ifd_count):
-            return (
-                struct.calcsize(tiff_header_struct)
-                + struct.calcsize(ifd_struct) * ifd_count
-                + 4
-            )
-
-        def add_ifd(tag_name, data, count=1):
-            key = tag_keys[tag_name]
-            typecode = TIFF_TAGS[key].type
-            ifds.append((key, typecode, count, data))
-
-        image_offset = None
-        add_ifd('ImageWidth', self.width)
-        add_ifd('ImageLength', self.height)
-        add_ifd('BitsPerSample', 1)
-        add_ifd('Compression', ccitt_group)
-        add_ifd('PhotometricInterpretation', int(photometry))
-        add_ifd('StripOffsets', lambda: image_offset)
-        add_ifd('RowsPerStrip', self.height)
-        add_ifd('StripByteCounts', img_size)
-
-        icc_offset = 0
-        if icc:
-            add_ifd('ICCProfile', lambda: icc_offset, count=len(icc))
-
-        icc_offset = header_length(len(ifds))
-        image_offset = icc_offset + len(icc)
-
-        ifd_args = [(arg() if callable(arg) else arg) for ifd in ifds for arg in ifd]
-        tiff_header = struct.pack(
-            (tiff_header_struct + ifd_struct[1:] * len(ifds) + 'L'),
-            b'II',  # Byte order indication: Little endian
-            42,  # Version number (always 42)
-            8,  # Offset to first IFD
-            len(ifds),  # Number of tags in IFD
-            *ifd_args,
-            0,  # Last IFD
+        return _transcoding.generate_ccitt_header(
+            self.size, img_size, ccitt_group, photometry, icc
         )
 
-        if icc:
-            tiff_header += icc
-
-        return tiff_header
-
     def show(self):  # pragma: no cover
-        """Show the image however PIL wants to"""
+        """Show the image however PIL wants to."""
         self.as_pil_image().show()
 
     def __repr__(self):
@@ -703,15 +771,17 @@ class PdfImage(PdfImageBase):
             f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
-    def _repr_png_(self):
-        """Display hook for IPython/Jupyter"""
+    def _repr_png_(self) -> bytes:
+        """Display hook for IPython/Jupyter."""
         b = BytesIO()
-        im = self.as_pil_image()
-        im.save(b, 'PNG')
-        return b.getvalue()
+        with self.as_pil_image() as im:
+            im.save(b, 'PNG')
+            return b.getvalue()
 
 
 class PdfJpxImage(PdfImage):
+    """Support class for JPEG 2000 images. Implements the same API as :class:`PdfImage`."""
+
     def __init__(self, obj):
         super().__init__(obj)
         self._jpxpil = self.as_pil_image()
@@ -725,10 +795,8 @@ class PdfJpxImage(PdfImage):
             and self._jpxpil == other._jpxpil
         )
 
-    def _extract_direct(self, *, stream):
-        data, filters = self._unstack_compression(
-            self.obj.get_raw_stream_buffer(), self.filters
-        )
+    def _extract_direct(self, *, stream: BinaryIO):
+        data, filters = self._remove_simple_filters(self.obj, self.filters)
         if filters != ['/JPXDecode']:
             raise UnsupportedImageTypeError(self.filters)
         stream.write(data)
@@ -748,7 +816,7 @@ class PdfJpxImage(PdfImage):
         raise NotImplementedError('Complex JP2 colorspace')
 
     @property
-    def _bpc(self):
+    def _bpc(self) -> int:
         # (PDF 1.7 Table 89) If the image stream uses the JPXDecode filter, this
         # entry is optional and shall be ignored if present. The bit depth is
         # determined by the conforming reader in the process of decoding the
@@ -756,7 +824,7 @@ class PdfJpxImage(PdfImage):
         return 8
 
     @property
-    def indexed(self):
+    def indexed(self) -> bool:
         # Nothing in the spec precludes an Indexed JPXDecode image, except for
         # the fact that doing so is madness. Let's assume it no one is that
         # insane.
@@ -770,7 +838,7 @@ class PdfJpxImage(PdfImage):
 
 
 class PdfInlineImage(PdfImageBase):
-    """Support class for PDF inline images"""
+    """Support class for PDF inline images. Implements the same API as :class:`PdfImage`."""
 
     # Inline images can contain abbreviations that we write automatically
     ABBREVS = {
@@ -794,7 +862,7 @@ class PdfInlineImage(PdfImageBase):
     }
     REVERSE_ABBREVS = {v: k for k, v in ABBREVS.items()}
 
-    def __init__(self, *, image_data, image_object: tuple):
+    def __init__(self, *, image_data: Object, image_object: tuple):
         """
         Args:
             image_data: data stream for image, extracted from content stream
@@ -817,7 +885,6 @@ class PdfInlineImage(PdfImageBase):
         except PdfError as e:
             raise PdfError("parsing inline " + reparse.decode('unicode_escape')) from e
         self.obj = reparsed_obj
-        self.pil = None
 
     def __eq__(self, other):
         if not isinstance(other, PdfImageBase):
@@ -850,7 +917,7 @@ class PdfInlineImage(PdfImageBase):
     def _metadata(self, name, type_, default):
         return metadata_from_obj(self.obj, name, type_, default)
 
-    def unparse(self):
+    def unparse(self) -> bytes:
         def metadata_tokens():
             for metadata_obj in self._image_object:
                 unparsed = self._unparse_obj(
@@ -869,7 +936,7 @@ class PdfInlineImage(PdfImageBase):
         return b''.join(inline_image_tokens())
 
     @property
-    def is_inline(self):
+    def is_inline(self) -> bool:
         return True
 
     @property
@@ -888,16 +955,30 @@ class PdfInlineImage(PdfImageBase):
             f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
-    def as_pil_image(self):
-        if self.pil:
-            return self.pil
+    def _convert_to_pdfimage(self):
+        # Construct a temporary PDF that holds this inline image, and...
+        tmppdf = Pdf.new()
+        tmppdf.add_blank_page(page_size=(self.width, self.height))
+        tmppdf.pages[0].contents_add(
+            f'{self.width} 0 0 {self.height} 0 0 cm'.encode('ascii'), prepend=True
+        )
+        tmppdf.pages[0].contents_add(self.unparse())
 
-        raise NotImplementedError('not yet')
+        # ...externalize it,
+        tmppdf.pages[0].externalize_inline_images()
+        raw_img = next(im for im in tmppdf.pages[0].images.values())
 
-    def extract_to(
-        self, *, stream=None, fileprefix=''
-    ):  # pylint: disable=unused-argument
-        raise UnsupportedImageTypeError("inline images don't support extract")
+        # ...then use the regular PdfImage API to extract it.
+        img = PdfImage(raw_img)
+        return img
+
+    def as_pil_image(self) -> Image.Image:
+        return self._convert_to_pdfimage().as_pil_image()
+
+    def extract_to(self, *, stream: Optional[BinaryIO] = None, fileprefix: str = ''):
+        return self._convert_to_pdfimage().extract_to(
+            stream=stream, fileprefix=fileprefix
+        )
 
     def read_bytes(self):
         raise NotImplementedError("qpdf returns compressed")
